"""Reinstallation utilities for accidental `pip uninstall`s."""

from .utils import get_valid_pip_history
from .utils import get_pip_command_action
from .utils import get_pip_command_packages


def get_reinstall_candidates():
    """Get candidates for reinstallation.

    Really, these aren't just candidates. There sure as h*ll the right ones.
    """
    pip_uninstall_history = [command for command in get_valid_pip_history()
                             if get_pip_command_action(command) == 'uninstall']
    assert pip_uninstall_history, 'Already pooped'
    return ' '.join(get_pip_command_packages(pip_uninstall_history[0]))



from colorama import init
from colorama import Style
from colorama import Fore
import os
import datetime
import pip
import platform
from subprocess import Popen
from subprocess import PIPE
from subprocess import STDOUT
from collections import defaultdict
from itertools import chain
import glob
import functools


init()
__all__ = ('get_package_groups', 'get_valid_pip_history', 'colored')


flatten = chain.from_iterable


def cache(f, history={}):
    @functools.wraps(f)
    def wrap(*args, **kwargs):
        if 'result' in history:
            return history['result']
        result = f(*args, **kwargs)
        history['result'] = result
        return result
    return wrap


def colored(string):
    """Print, coloring all code delimited with backticks."""
    count = string.count('`') // 2
    args = [Fore.GREEN + Style.BRIGHT, Style.RESET_ALL] * count
    return string.replace('`', '{}').format(*args)


def get_package_groups():
    """Get groups of packages installed at the same time.

    The list is ordered in reverse order, across time as (time, packages). Here,
    time is formatted as datetime objects.
    """
    mapping = defaultdict(lambda: [])
    for package_name_version, creation_time in get_packages_creation_times():
        mapping[creation_time].append(package_name_version)

    result = []
    for creation_time in sorted(mapping, reverse=True):
        pretty_mod_time = datetime.datetime.fromtimestamp(creation_time)
        result.append((pretty_mod_time, mapping[creation_time]))
    return result


@cache
def get_valid_pip_history(num_bash_entries=10):
    """Looks only at valid pip commands logged in history.

    See `is_valid_command` for details on `valid` commands.
    """
    commands = get_pip_history(num_bash_entries=num_bash_entries)
    for command in commands:
        if is_invalid_command(command):
            commands.remove(command)
    return commands


def is_invalid_command(command):
    """Check if the given command is valid.

    Details on what constitutes a `valid` command:
    - install: If none of the packages are still installed, ignore it
    - uninstall: If all of the packages are installed, ignore it
    - download: If no tars or wheels are found in the current directory,
                related to the specified packages, ignore it

    Any other pip command is unsupported, as it does not make sense to undo
    those functions.
    """
    all_packages = set(get_pip_package_key(package) for package
                       in pip.get_installed_distributions())
    action = get_pip_command_action(command)
    packages = get_pip_command_packages(command)
    packages = set([p.split('==')[0] for p in packages])
    return (action == 'install' and not all_packages & packages) or \
        (action == 'uninstall' and not packages - all_packages) or \
        (action == 'download' and not any(
            [get_associated_file(package) for package in packages])) or \
        (action not in ('install', 'uninstall', 'download'))


def get_associated_file(package_key):
    """Find file associated with the package key."""
    package_key = package_key.replace('-', '_')
    for path in flatten([glob.iglob('*.whl'), glob.iglob('*.tar.gz')]):
        if path.startswith(package_key):  # TODO: more specific format check
            return path


def get_pip_package_key(candidate):
    """Get package key for the provided string.

    The string may be a name and version.
    """
    return str(candidate).split()[0].split('==')[0]


def get_pip_command_action(command):
    """Return pip action for a pip command."""
    return command.split()[1]


def get_pip_command_packages(command):
    """Return packages included in a pip command."""
    return command.split()[2:]


def get_pip_history(num_bash_entries=10):
    """Get all pip commands logged in history."""
    commands = []
    for line in get_bash_history(num_bash_entries=num_bash_entries):
        for command in line.split(';'):
            if command.strip().startswith('pip'):
                commands.append(command)
    return commands


@cache
def get_bash_history(num_bash_entries=10):
    """Get bash history.

    Items are sorted in reverse order, starting with the most recent.
    """
    n = '' if num_bash_entries == '*' else str(num_bash_entries)
    shell_command = 'bash -i -c "history -r; history %s"' % n
    event = Popen(shell_command, shell=True, stdin=PIPE, stdout=PIPE,
                  stderr=STDOUT)
    output, error = event.communicate()
    output = output.decode('utf-8').replace("'\n", '\n')  # clean up weirdness
    commands = [' '.join(str(line).split()[1:]) for line in output.splitlines()]
    return commands[::-1]


def get_packages_creation_times():
    """Get all pip packages and their timestamps.

    Original function taken from StackOverflow:
    https://stackoverflow.com/a/44436961/4855984
    """
    packages = []
    for package in pip.get_installed_distributions():
        package_name_version = str(package)
        package_location = get_package_location(package)
        creation_time = get_creation_date(package_location)
        packages.append([
            package_name_version,
            creation_time,
        ])
    return packages


def get_package_location(package):
    """Get module directory from the module information.

    Original function taken from StackOverflow:
    https://stackoverflow.com/a/44436961/4855984
    """
    try:
        module_dir = next(package._get_metadata('top_level.txt'))
        package_location = os.path.join(package.location, module_dir)
        os.stat(package_location)
        return package_location
    except (StopIteration, OSError):
        pass
    package_location = os.path.join(package.location, package.key)
    if os.path.isdir(package_location):
        return package_location
    return package.location


def get_creation_date(path_to_file):
    """
    Try to get the date that a file was created, falling back to when it was
    last modified if that isn't possible.
    See http://stackoverflow.com/a/39501288/1709587 for explanation.
    """
    if platform.system() == 'Windows':
        return os.path.getctime(path_to_file)
    else:
        stat = os.stat(path_to_file)
        try:
            return stat.st_birthtime
        except AttributeError:
            # We're probably on Linux. No easy way to get creation dates here,
            # so we'll settle for when its content was last modified.
            return stat.st_mtime



"""Uninstallation utilities for accidental `pip install`s"""

from .utils import get_valid_pip_history
from .utils import get_package_groups
from .utils import get_pip_command_action
from .utils import get_pip_command_packages
from .utils import get_pip_package_key
from .utils import flatten
from pipdeptree import build_dist_index
from pipdeptree import construct_tree
from pipdeptree import reverse_tree
from pipdeptree import find_tree_root
import pip


def get_installed_package_keys():
    """Get all installed package keys."""
    return [get_pip_package_key(package) for package in
            pip.get_installed_distributions()]


def get_most_recent_packages_by_creation_time():
    """Get most recent packages as determined by module source create time."""
    package_groups = get_package_groups()
    if not package_groups:
        return set()
    time, packages = package_groups[0]
    packages = [pkg.split()[0] for pkg in packages]
    return set(packages)


def get_most_recent_packages_by_pip_history():
    """Get most recent packages as determined by pip history."""
    pip_install_history = [command for command in get_valid_pip_history()
                           if get_pip_command_action(command) == 'install']
    if not pip_install_history:
        return set()
    packages = [package.split('==')[0] for package in
                get_pip_command_packages(pip_install_history[0])]
    return set(packages)


def get_uninstall_dependencies_for(package_keys):
    """Find candidates for installation given package keys.

    1. Find all dependencies for these packages.
    2. Determine which dependencies:
       a. are not needed by any other package and
       b. have a "recent" creation date.
    """
    # 1. Find all dependencies of these packages.
    all_packages = pip.get_installed_distributions()
    dist_index = build_dist_index(all_packages)
    packages = set(dist_index[package_key] for package_key in package_keys)
    dependencies = flatten([package.requires() for package in packages])

    # 2. Determine unneeded dependencies that were installed at roughly the
    #    same time. # TODO: check time of installation
    candidates = set(package_keys)
    rtree = reverse_tree(construct_tree(dist_index))
    for dep in dependencies:
        node = find_tree_root(rtree, dep.key)
        leftover = {d.key for d in rtree[node] if d.key not in package_keys}
        if not leftover:
            candidates.add(dep.key)
    return ' '.join(candidates)


def get_uninstall_candidates():
    """Find candidates for uninstallation.

    1. Find recent modules by creation time and history. Merge the two.
    2. Find all relevant dependencies for uninstallation.
    """
    # 1. Find packages that were recently installed.
    packages_by_module = get_most_recent_packages_by_creation_time()
    packages_by_history = get_most_recent_packages_by_pip_history()
    package_keys = packages_by_module.union(packages_by_history)
    all_packages = set(get_installed_package_keys())
    package_keys = [p for p in package_keys if p in all_packages]
    assert package_keys, 'Possible uninstall candidates have' \
                         ' already been uninstalled.'

    # 2. Determine relevant dependencies.
    return get_uninstall_dependencies_for(package_keys)


import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand


tests_require = ['pytest', 'pytest-cov==2.5.1', 'coverage == 3.7.1',
                 'coveralls == 1.1']
install_requires = ['pipdeptree==0.10.1', 'colorama==0.3.9']


class PyTest(TestCommand):

    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


VERSION = '0.1.1'


setup(
    name="LePoop",
    version=VERSION,
    author="Alvin Wan",
    author_email='hi@alvinwan.com',
    description="Allows you to undo pip commands",
    license="BSD",
    url="https://github.com/alvinwan/lepoop",
    packages=['lepoop'],
    cmdclass={
        'test': PyTest,
    },
    tests_require=tests_require,
    install_requires=install_requires + tests_require,
    download_url='https://github.com/alvinwan/lepoop/archive/%s.zip' % VERSION,
    classifiers=[
        "Topic :: Utilities",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
    ],
    zip_safe=False,
    entry_points={'console_scripts': [
        'lepoop = lepoop.entry.main:main',
        'poop = lepoop.entry.not_configured:main'
    ]}
)



"""Utilities for reversing download."""

from .utils import get_valid_pip_history
from .utils import get_pip_command_action
from subprocess import Popen
from subprocess import PIPE
from subprocess import STDOUT


def get_file_candidates(prefix='File was already downloaded'):
    """Find all files likely to be related to the latest packages."""
    pip_download_history = [command for command in get_valid_pip_history()
                            if get_pip_command_action(command) == 'download']
    assert pip_download_history, 'Already pooped'
    event = Popen(pip_download_history[0].split(), stdin=PIPE,
                  stdout=PIPE, stderr=STDOUT)
    output, error = event.communicate()

    # Search output for files
    candidates = []
    for line in output.decode('utf-8').splitlines():
        if line.strip().startswith(prefix):
            filename = line.replace(prefix, '').strip()
            candidates.append(filename)
    return ' '.join(candidates)
